/**
 *  Copyright (c) 2014, 
 *  BTC Embedded Systems AG, Oldenburg, Germany  
 *  All rights reserved
 *  --------------------------------------------------------- 
 *  $Author: $
 *  $Date:  $
 *  $Revision: $
 **/

package com.btc.ep.base.dal.tests.it;

import java.util.ArrayList;
import java.util.List;

import org.junit.After;
import org.junit.AfterClass;
import org.junit.Before;
import org.junit.BeforeClass;
import org.junit.FixMethodOrder;
import org.junit.Test;
import org.junit.runners.MethodSorters;

import com.btc.ep.base.bl.Vector;
import com.btc.ep.base.bl.VectorElement;
import com.btc.ep.base.bl.impl.VectorElementImpl;
import com.btc.ep.base.bl.impl.VectorImpl;
import com.btc.ep.base.bl.repository.VectorRepository;
import com.btc.ep.base.dal.tests.it.base.BaseTest;
import com.btc.ep.base.transactions.TransactionManager;
import com.btc.ep.core.Services;

/**
 * This class tests all methods of the message repository.
 * 
 * @author Simon Kollmann
 * @see VectorRepository
 */
@FixMethodOrder(MethodSorters.NAME_ASCENDING)
public class PerformanceTest extends BaseTest {

	private static final long STEPS = 100000;

	/**
	 * initializes resources before running all tests
	 * 
	 * @throws Exception
	 *             anything
	 */
	@BeforeClass
	public static void setUpBeforeClass() throws Exception {

	}

	
	/**
	 * Create Vector   SLOW
	 */
	// @Transactional
	private static Vector createVector_OneTransaction(String name, int steps,
			int stepsPerCollection) {

		TransactionManager tm = Services.get(TransactionManager.class);
		tm.init(jointPointId, false);

		long startTime = System.currentTimeMillis();

		Vector vector = persistVector(name);

		for (int m = 0; m < steps / stepsPerCollection; m++) {

			createColletion(vector.getUid(), m, stepsPerCollection);
		}

		System.out.println(" create Vector " + name + " complete");
		System.out.println("vector " + name + " create time: "
				+ (System.currentTimeMillis() - startTime));

		tm.commit(jointPointId);
		tm.close(jointPointId);

		return findVector(vector.getUid());
	}
	
	/**
	 * Persist Vector   FAST
	 */
	// @NotTransactional
	private static Vector createVector(String name, int steps,
			int stepsPerCollection) {

		long startTime = System.currentTimeMillis();

		Vector vector = persistVector(name);

		for (int m = 0; m < steps / stepsPerCollection; m++) {

			createColletion(vector.getUid(), m, stepsPerCollection);
		}

		System.out.println(" create Vector " + name
				+ " complete NotTransactional");
		System.out.println("vector " + name + " create time: "
				+ (System.currentTimeMillis() - startTime));

		return findVector(vector.getUid());
	}



	/**
	 * destroys resources after running all tests
	 * 
	 * @throws Exception
	 *             anything
	 */
	@AfterClass
	public static void tearDownAfterClass() throws Exception {

	}

	/**
	 * initializes before running a particular test
	 * 
	 * @throws Exception
	 *             anything
	 */
	@Before
	public void setUp() throws Exception {

		deleteDatabase();
	}

	/**
	 * destroys after running a particular test
	 * 
	 * @throws Exception
	 *             anything
	 */
	@After
	public void tearDown() throws Exception {
	}

	@Test
	public void test1_OneTransaction() {
		System.out.println("test1_OneTransaction");

		deleteDatabase();

		//SLOW
		String name = "vectorA";
		int steps = (int) STEPS;
		int stepsPerCollection = 10000;
		
		TransactionManager tm = Services.get(TransactionManager.class);
		tm.init(jointPointId, false);

		long startTime = System.currentTimeMillis();

		Vector vector = persistVector(name);

		for (int m = 0; m < steps / stepsPerCollection; m++) {

			createColletion(vector.getUid(), m, stepsPerCollection);
		}

		System.out.println(" create Vector " + name + " complete");
		System.out.println("vector " + name + " create time: "
				+ (System.currentTimeMillis() - startTime));

		tm.commit(jointPointId);
		tm.close(jointPointId);

	}

	@Test
	public void test2_ManyTransactions() {
		System.out.println("test2_ManyTransactions");

		deleteDatabase();

		//FAST
		String name = "vectorA";
		int steps = (int) STEPS;
		int stepsPerCollection = 10000;
		
		long startTime = System.currentTimeMillis();

		Vector vector = persistVector(name);

		for (int m = 0; m < steps / stepsPerCollection; m++) {

			createColletion(vector.getUid(), m, stepsPerCollection);
		}

		System.out.println(" create Vector " + name
				+ " complete NotTransactional");
		System.out.println("vector " + name + " create time: "
				+ (System.currentTimeMillis() - startTime));
	}

	
	
	private static void createColletion(String vectorID, int collectionNr,
			int stepsPerCollection) {
		TransactionManager tm = Services.get(TransactionManager.class);
		tm.init(jointPointId, false);

		VectorRepository vectorRepository = Services
				.get(VectorRepository.class);
		Vector vector = vectorRepository.findVector(vectorID);

		List<VectorElement> vectorElements = new ArrayList<VectorElement>(
				stepsPerCollection);
		for (int i = 0; i < stepsPerCollection; i++) {
			long step = (collectionNr * stepsPerCollection) + i;

			VectorElementImpl ve = new VectorElementImpl(step,
					createFloatArray(step));
			vectorElements.add(ve);
		}
		vectorRepository.addVectorElements(vector, vectorElements);
		vector.incrementElementCount(vectorElements.size());

		// ?? whitout, it doesn't work
		vectorRepository.persistVector(vector);

		tm.commit(jointPointId);
		tm.close(jointPointId);
	}
	
	/**
	 * Persist Vector
	 */
	private static Vector persistVector(String name) {
		TransactionManager tm = Services.get(TransactionManager.class);
		tm.init(jointPointId, false);
		
		VectorRepository vectorRepository = Services
				.get(VectorRepository.class);
		Vector vector = new VectorImpl();
		vector.setName("name");
		vector = vectorRepository.persistVector(vector);
		
		tm.commit(jointPointId);
		tm.close(jointPointId);
		return vector;
	}


	private static List<Float> createFloatArray(float value) {
		List<Float> floats = new ArrayList<Float>(100);
		for (float i = 0; i < 100; i++) {
			floats.add(value + (i / 100));
		}
		return floats;
	}

	private static Vector findVector(String uid) {
		TransactionManager tm = Services.get(TransactionManager.class);
		tm.init(jointPointId, false);

		VectorRepository vectorRepository = Services
				.get(VectorRepository.class);
		Vector vector = vectorRepository.findVector(uid);
		
		tm.commit(jointPointId);
		tm.close(jointPointId);
		
		return vector;
	}
}
