/**
 *  Copyright (c) 2014, 
 *  BTC Embedded Systems AG, Oldenburg, Germany  
 *  All rights reserved
 *  --------------------------------------------------------- 
 *  $Author: $
 *  $Date:  $
 *  $Revision: $
 **/

package com.btc.ep.base.dal.internal;

import java.util.ArrayList;
import java.util.List;

import com.btc.ep.base.bl.Vector;
import com.btc.ep.base.bl.VectorElement;
import com.btc.ep.base.bl.impl.VectorElementCollectionImpl;
import com.btc.ep.base.bl.impl.VectorElementImpl;
import com.btc.ep.base.bl.impl.VectorImpl;
import com.btc.ep.base.bl.repository.VectorRepository;
import com.btc.ep.base.dal.impl.BaseRepository;
import com.btc.ep.dal.persist.PersistenceManager;
import com.btc.ep.dal.persist.impl.PersistenceService;

/**
 * This class is a the Repository (Service) to access Messages.
 * 
 * @author Simon Kollmann
 */
public class VectorRepositoryImpl implements VectorRepository {

    /**
     * Injected by OSGi.
     */
    private PersistenceManager persistenceManager;

    /**
     * Injected by OSGi.
     */
    private BaseRepository<Vector> vectorRepository;
    private BaseRepository<VectorElement> vectorElementRepository;
    private BaseRepository<VectorElementCollectionImpl> vectorElementCollectionRepository;

    /**
     * Set the PersistenceManager
     *
     * @param persistenceManager The PersistenceManager
     */
    public void setPersistenceManager(PersistenceManager persistenceManager) {
        this.persistenceManager = persistenceManager;
    }

    /**
     * Set the VectorRepository
     *
     * @param repository The VectorRepository
     */
    public void setVectorRepository(BaseRepository<Vector> repository) {
        this.vectorRepository = repository;
    }

    /**
     * Set the VectorElementRepository
     *
     * @param repository The VectorElementRepository
     */
    public void setVectorElementRepository(BaseRepository<VectorElement> repository) {
        this.vectorElementRepository = repository;
    }

    /**
     * Set the VectorElementCollectionRepository
     *
     * @param repository The VectorElementCollectionRepository
     */
    public void setVectorElementCollectionRepository(BaseRepository<VectorElementCollectionImpl> repository) {
        this.vectorElementCollectionRepository = repository;
    }

    @Override
    public Vector findVector(String uid) {
    	PersistenceService persistenceService = persistenceManager.getPersistenceService();
    	return persistenceService.find(VectorImpl.class, uid);
    }

    @Override
    public List<Vector> findAllVectors() {
        return new ArrayList<Vector>(vectorRepository.findAll(VectorImpl.class));
    }

    @Override
    public Vector persistVector(Vector vector) {
        return vectorRepository.update(vector);
    }

    @Override
    public void removeVector(Vector vector) {

        // remove the VectorElements
        List<VectorElement> vectorElements = findVectorElements(vector);
        vectorElementRepository.remove(vectorElements);

        // remove the VectorElementCollections
        List<VectorElementCollectionImpl> vectorElementCollections = findVectorElementCollections(vector);
        vectorElementCollectionRepository.remove(vectorElementCollections);

        //remove the vector
        vectorRepository.remove(vector);
    }

    /*
     * VectorElemt(s)
     */

    @Override
    public VectorElement findVectorElement(String uid) {
        return vectorElementRepository.find(uid);
    }

    @Override
    public List<VectorElement> findVectorElements(Vector vector) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        //TOTO extract
        List<Long> vectorElemmentCollectionRefs =
            persistenceService.getModelElementsByQuery("select vec.vectorElementCollectionRef from "
                + VectorElementCollectionImpl.class.getSimpleName() + " vec where vec.vector = ?1", long.class, vector);

        return persistenceService.getModelElementsByQuery("select ve from " + VectorElementImpl.class.getSimpleName()
            + " ve where ve.vectorElementCollectionRef MEMBER OF ?1 ORDER BY ve.stepOffset ASC", VectorElement.class,
            vectorElemmentCollectionRefs);

        //FIXME fix the step number's
        //TODO fix the step number's

    }

    @Override
    public List<VectorElement> findVectorElements(Vector vector, long startStep, long endStep) {
        // TODO Auto-generated method stub
        throw new UnsupportedOperationException("Auto-generated method stub");
    }

    @Override
    public VectorElement findVectorElement(Vector vector, long step) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        VectorElementCollectionImpl vec;
        vec = findVectorElementCollection(vector, step);

        long stepOffset = step - vec.getStartStep();

        VectorElement ve =
            persistenceService.getModelElementOrNullByQuery("select ve from " + VectorElementImpl.class.getSimpleName()
                + " ve where ve.vectorElementCollectionRef = ?1 AND ve.stepOffset = ?2", VectorElement.class,
                vec.getVectorElementCollectionRef(), stepOffset);

        ve.setStep(ve.getStepOffset() + vec.getStartStep());

        return ve;
    }

    @Override
    public void incrAll() {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        //Increment (all) steps

        persistenceService.updateQuery("UPDATE " + VectorElementImpl.class.getSimpleName()
            + " AS ve SET ve.stepOffset = ve.stepOffset + 1");
    }

    private VectorElementCollectionImpl findVectorElementCollection(Vector vector, long step) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();
        long start_findVectorElementCollection = System.currentTimeMillis();
        
        VectorElementCollectionImpl vec =
            persistenceService.getModelElementOrNullByQuery(
                "select vec from " + VectorElementCollectionImpl.class.getSimpleName()
                    + " vec where vec.vector = ?1 AND vec.startStep <= ?2 AND ?2 <= vec.endStep",
                VectorElementCollectionImpl.class, vector, step);
        
        System.out.println("findVectorElementCollection  time: " + (System.currentTimeMillis() - start_findVectorElementCollection));
        return vec;
    }

    private List<VectorElementCollectionImpl> findVectorElementCollections(Vector vector) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        List<VectorElementCollectionImpl> vecs =
            persistenceService.getModelElementsByQuery(
                "select vec from " + VectorElementCollectionImpl.class.getSimpleName() + " vec where vec.vector = ?1"
                    + " ORDER BY vec.startStep ASC", VectorElementCollectionImpl.class, vector);
        return vecs;
    }

    private List<VectorElementCollectionImpl> findVectorElementCollections(Vector vector, long startStep, long endStep) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        List<VectorElementCollectionImpl> vecs =
            persistenceService.getModelElementsByQuery(
                "select vec from " + VectorElementCollectionImpl.class.getSimpleName()
                    + " vec where vec.vector = ?1 AND (" + " (vec.startStep <= ?2 AND ?2 <= vec.endStep) OR"
                    + " (?2 <= vec.startStep AND vec.endStep <= ?3) OR"
                    + " (vec.startStep <= ?3 AND ?3 <= vec.endStep ) )" + " ORDER BY vec.startStep ASC",
                VectorElementCollectionImpl.class, vector, startStep, endStep);
        return vecs;
    }

    /**
     * IncrementSteps
     *
     * @param vector
     * @param startStep
     * @param offset
     * @return VectorElementCollection which now has a gap.
     */
    private VectorElementCollectionImpl incrementSteps(Vector vector, long startStep, long offset) {
        long start_incrementSteps = System.currentTimeMillis();
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        boolean vecIsTheLast = false;
        //pre check
        Long size = ((VectorImpl)vector).getElementCount();

        if (size < startStep) {
            throw new RuntimeException(" incrementSteps was not possible (size " + size + " < startStep " + startStep
                + ")");
        }

        // find VectorElementCollection
        VectorElementCollectionImpl vec;
        if (size > startStep) {
            vec = findVectorElementCollection(vector, startStep);
        } else {
            vecIsTheLast = true;
            // find the last VectorElementCollection of the vector
            vec = findVectorElementCollection(vector, size - 1);
        }

        if (vec == null) {
            //create the first collection with correct size
            vec = new VectorElementCollectionImpl(vector, nextVectorElementCollectionRef(), 0, offset - 1);
            vec = vectorElementCollectionRepository.update(vec);
        } else {

            //Increment steps of the VectorElementCollection, starting by startStep
            if (!vecIsTheLast) {
                long startStepOffset = startStep - vec.getStartStep();
                // UPDATE VectorElementImpl AS ve SET ve.stepOffset = ve.stepOffset + 1 WHERE ve.vectorElementCollectionRef = 1 AND ve.stepOffset >= 500
                persistenceService.updateQuery("UPDATE " + VectorElementImpl.class.getSimpleName()
                    + " AS ve SET ve.stepOffset = ve.stepOffset + ?3"
                    + " WHERE ve.vectorElementCollectionRef = ?1 AND ve.stepOffset >= ?2",
                    vec.getVectorElementCollectionRef(), startStepOffset, offset);
            }
            // Increment the size of VectorElementCollection
            vec.incrementSize(offset);
            //        //UPDATE  VectorElementCollectionImpl AS vec SET vec.endStep = vec.endStep + 1    WHERE vec.vectorElementCollectionRef = 1
            //        persistenceService.updateQuery("UPDATE " + VectorElementCollectionImpl.class.getSimpleName()
            //            + " AS vec SET vec.endStep = vec.endStep + 1" + " WHERE vec = ?1", vec);

            if (!vecIsTheLast) {
                // Increment the start and end step of all other vec with vec.startStep > startStep
                //UPDATE VectorElementCollectionImpl AS vec SET vec.startStep = vec.startStep + 1, vec.endStep = vec.endStep + 1 WHERE vec.startStep > 500
                persistenceService.updateQuery("UPDATE " + VectorElementCollectionImpl.class.getSimpleName()
                    + " AS vec SET vec.startStep = vec.startStep + ?3, vec.endStep = vec.endStep + 1"
                    + " WHERE vec.vector = ?1 AND vec.startStep > ?2", vector, startStep, offset);
            }
        }
        System.out.println("incrementSteps  time: " + (System.currentTimeMillis() - start_incrementSteps));
        return vec;
    }

    /**
     * decrementSteps
     * 
     * First remove the Elements, then decrement the steps.
     *
     * @param vector
     * @param startStep
     * @param endStep
     */
    private void decrementSteps(Vector vector, long startStep, long endStep) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();

        //pre check
        Long size = ((VectorImpl)vector).getElementCount();

        if (startStep > size) {
            throw new RuntimeException(" decrementSteps was not possible ( startStep " + startStep + " > size " + size
                + ")");
        }

        // find VectorElementCollections
        List<VectorElementCollectionImpl> vectorElementCollections;
        vectorElementCollections = findVectorElementCollections(vector, startStep, endStep);

        long removedElements = endStep - startStep + 1;

        //decrement the steps
        for (VectorElementCollectionImpl vec : vectorElementCollections) {

            if (startStep <= vec.getStartStep() && vec.getEndStep() <= endStep) {
                //all elements between startStep and endStep

                //VectorElementCollection contains after "decrement" no Elements
                //-> remove the VectorElementCollection
                vectorElementCollectionRepository.remove(vec);
            } else if (startStep > vec.getStartStep() && vec.getEndStep() <= endStep) {
                //ab mitte bis ende

                //change the end Value
                vec.setEndStep(startStep - 1);

            } else if (vec.getStartStep() <= endStep && endStep < vec.getStartStep()) {
                // x bis mitte

                //Decrement steps of the VectorElementCollection, starting by startStepOffset

                //Offset Value to start the adjustment
                long startStepOffset = 0;
                long decrement = endStep - vec.getStartStep() + 1;
                if (startStep > vec.getStartStep()) {
                    startStepOffset = startStep - vec.getStartStep();
                    decrement = removedElements;
                }

                // UPDATE VectorElementImpl AS ve SET ve.stepOffset = ve.stepOffset + 1 WHERE ve.vectorElementCollectionRef = 1 AND ve.stepOffset >= 500
                persistenceService.updateQuery("UPDATE " + VectorElementImpl.class.getSimpleName()
                    + " AS ve SET ve.stepOffset = ve.stepOffset - ?3"
                    + " WHERE ve.vectorElementCollectionRef = ?1 AND ve.stepOffset >= ?2",
                    vec.getVectorElementCollectionRef(), startStepOffset, decrement);

                //change the start Value
                if (startStep <= vec.getStartStep()) {
                    // vec.getStartStep() - (vec.getStartStep() - startStep)  ==  startStep
                    vec.setStartStep(startStep);
                }

                //change the end Value
                vec.setEndStep(vec.getEndStep() - removedElements);

            } else if (endStep < vec.getStartStep()) {
                //after endStep
                //all steps must decrement
                vec.decrementSteps(removedElements);
            } else {
                //(vec.getEndStep() < startStep)
                //before startStep
                // do noting
            }
        }
    }

    @Override
    public VectorElement addVectorElement(Vector vector, VectorElement vectorElement) {

        //Pre check
        Long size = ((VectorImpl)vector).getElementCount();
        if (size < vectorElement.getStep()) {
            //throw new RuntimeException("new vector element has a non-matching step.");
            System.out.println("ERROR:  new vector element has a non-matching step.");
        }

        VectorElementCollectionImpl vec;
        vec = incrementSteps(vector, vectorElement.getStep(), 1);

        // set the VectorElementCollection Reference
        vectorElement.setVectorElementCollectionRef(vec.getVectorElementCollectionRef());

        //persist the VectorElement
        return vectorElementRepository.update(vectorElement);

        //TODO check the size of the vec
    }

    @Override
    public List<VectorElement> addVectorElements(Vector vector, List<VectorElement> steps) {

        long start_addVectorElements = System.currentTimeMillis();

        long startStep = steps.get(0).getStep();

        //Pre check
        Long size = ((VectorImpl)vector).getElementCount();
        if (size < startStep) {
            //throw new RuntimeException("new vector element has a non-matching step.");
            System.out.println("ERROR:  new vector element has a non-matching step.");
        }

        //TODO check that the steps are ascending!

        VectorElementCollectionImpl vec;
        vec = incrementSteps(vector, startStep, steps.size());

        // set the vectorElementCollection reference
        for (VectorElement vectorElement : steps) {
            vectorElement.setVectorElementCollectionRef(vec.getVectorElementCollectionRef());
        }

        //persist the VectorElements
        long start_persist = System.currentTimeMillis();
        ArrayList<VectorElement> vectorElements = new ArrayList<VectorElement>(vectorElementRepository.update(steps));
        System.out.println("addVectorElements: persist  time: " + (System.currentTimeMillis() - start_persist));

        
        
        System.out.println("addVectorElements total time: " + (System.currentTimeMillis() - start_addVectorElements)
            + "\n");

        return vectorElements;

        //TODO check the size of the vec
    }

    /**
     * Determine the number of existing VectorElements of a vector
     *
     * @param vector Vector
     * @return number of existing VectorElements
     */
    private long countElementsOfVector(Vector vector) {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();
        //find count of VectorElements
        Long size =
            persistenceService.getSingleResultByQuery(
                "select vector.elementCount from " + VectorImpl.class.getSimpleName() + " vector WHERE vector = ?1",
                long.class, vector);
        return size;
    }

    /**
     * Determine the next number for VectorElementCollection Reference
     *
     * @return number of next VectorElementCollection Reference
     */
    private long nextVectorElementCollectionRef() {
        PersistenceService persistenceService = persistenceManager.getPersistenceService();
        Long size =
            persistenceService.getSingleResultByQuery("select MAX(vec.vectorElementCollectionRef) from "
                + VectorElementCollectionImpl.class.getSimpleName() + " vec", long.class);

        if (size == null) {
            return 0;
        }

        return size + 1;
    }

    @Override
    public void removeVectorElement(Vector vector, long step) {

        //find the Element
        VectorElement ve = findVectorElement(vector, step);

        removeVectorElement(vector, ve);
    }

    @Override
    public void removeVectorElement(Vector vector, VectorElement vectorElement) {

        long step = vectorElement.getStep();
        //remove
        vectorElementRepository.remove(vectorElement);

        decrementSteps(vector, step, 1);
    }

    @Override
    public void removeVectorElements(Vector vector, long startStep, long endStep) {

        List<VectorElement> vectorElements = findVectorElements(vector, startStep, endStep);
        vectorElementRepository.remove(vectorElements);

        decrementSteps(vector, startStep, endStep);
    }
}
